# Introduction

This document provides an overview of the TIBCO Track and Trace Accelerator.  

# Overview

The Track and Trace Accelerator is a fully cloud-native application that runs on products in TIBCO Cloud. It captures data feeds related to the lifecycle of a parcel as it transits from collection point through to delivery point. Proactive SLA monitoring recalculates whether or not a parcel will be delivered in time and publishes alerts and notifications in the event of SLA violation. Parcel containerization allows for rolling up parcel events into containers and gives visibility of affected packages in the event of container delays.

## Business Scenario

The logistics industry has experienced a great deal of change over the past decade. Traditional postal products are dropping off the radar, giving a new importance to parcel delivery. The proliferation of eCommerce and online shopping has made efficient and timely deliveries key to both vendor and customer satisfaction. Retailers entering the logistics market like Amazon and Ocado have increased pressure on incumbent operators to improve their offerings. These new entrants are not necessarily better, but they can be cheaper and more agile, combined with modern information systems help them to be more efficient. But customers ultimately drive revenue, and if they're unhappy with a delivery they will complain to the retailer. Retailers want happy customers, and a healthy logistics marketplace is required to improve services and maintain competitive pricing.

## Concepts

The Accelerator is an in-memory model of a parcel delivery lifecycle. It receives two types of external events which are used to build an internal state model.

* **Announcements** -- this is the initial notification that a parcel is ready for collection
* **Observations** -- this represents a barcode scan at various points during the delivery lifecycle

## Benefits and Business Value

The Track and Trace Accelerator is a reference architecture for a new generation tracking technology based on event processing software. The process is configuration-driven which allows for easy modification of states and allows introduction of new logistics products in days or hours rather than weeks. The real-time dashboard for operations staff gives an at-a-glance view of network health while flagging up exceptions that require attention. With active SLA management the system can predict when parcels will be late, upwards of 8 hours before this actually happens. This also provides tools for adaptable last mile delivery changes based on customer preference.

## Technical Scenario

The Accelerator shows several use cases tracking a package throughout it lifecycle from collection through to delivery. The parcel is announced to the system, and then a series of barcode scans known as observations occur as the parcel transits the network. Updates on estimated delivery are recalculated at every step so that active monitoring of SLA violations can determine if a parcel will be delivered within the guidelines for a given product. 

## Components

The main business rules engine is implemented using TIBCO Cloud Events as the Event Manager. This captures dynamic events and applies business logic to classify the wireline state.

The Operational Dashboard is implemented using TIBCO Cloud Data Streams and TIBCO Spotfire®. It captures the current state of the parcel delivery network. It displays this information on a fully-interactive, TIBCO Spotfire® Web Player dashboard which can be customized by users to their needs.

![](./doc/images/component.png)

# Installation

Before installing the Accelerator, the user will need to decide on a deployment environment:

Environment | Description
------- | -------
[TIBCO Cloud](./doc/cloud-getting-started.md) | Suitable for demo and runtime uses. Requires TIBCO Cloud account with enabled subscriptions.

For further instructions how to setup these environments select the appropriate link above.
