# Introduction

This document will guide you in setting up the Accelerator on TIBCO Cloud. This requires access to a TIBCO Cloud account with services enabled.

After completing the setup, you will have an environment suitable for demo and runtime uses. 

# Architecture

This is what you will be provisioning:

![](./images/cloud/cloud-architecture.png)

# Pre-Requisites

## TIBCO Cloud Requirements

For TIBCO Cloud, you will require an account with access to the following enabled subscriptions:

* TIBCO Cloud Data Streams (with either User or Administrator privileges)
* TIBCO Cloud Events
* TIBCO Cloud Integration
* TIBCO Cloud Messaging
* TIBCO Cloud Spotfire

At the moment only installations in the TIBCO Cloud EU or US regions are supported. All subscriptions must be enabled in the same region.

## TIBCO Cloud Data Streams Notes

1. You must have either User or Administrator privileges for your TIBCO Cloud Data Streams subscription in order to be able to create streams. If you do not have these privileges then the automated provisioning will fail.

1. Most TIBCO Cloud Data Streams subscriptions have a limit to the number of simultaneous running streams. For example, 5 streams is a common configuration. if you have running streams in your subscription, ensure you have at least 1 available stream to start for the Accelerator. If there is not 1 available stream then the automated provisioning will fail.

# Download Accelerator Distribution

The Accelerator is distributed as two packages:

* [Distribution](https://download.streambase.com/serve/Accelerators/TrackTrace_dist_1.0.2.zip) -- distribution for building the application</li>
* [Source](https://download.streambase.com/serve/Accelerators/TrackTrace_src_1.0.2.zip) -- to view and make changes to the source</li>

Download the **distribution** package to your local machine and extract the ZIP into a temporary directory.

![](./images/cloud/extracted-zip.png)

Once extracted you will have a directory called **TrackTrace** containing the files **manifest.json** and **track-webapp.zip**. 

# Deploy Accelerator Webapp

To deploy the Accelerator webapp you will need to login to your TIBCO Cloud Integration account. Click [here for EU region](https://eu.integration.cloud.tibco.com/) and [here for US region](https://integration.cloud.tibco.com/). You may be prompted to login using your TIBCO Cloud credentials.

Click the **Create/Import** button, and then choose a **Node.js** application under the **Develop** section.

![](./images/cloud/create-import-nodejs.png)

Click the **Next** button. Drag and drop both the **manifest.json** and **track-webapp.zip** into the displayed dialog, then click **Import App**. The **Creating a new application..** message will display for some time while the app is created.

Once the app is successfully created, the following page will be displayed.

![New Node App](./images/cloud/new-node-app.png)

To start the app, hover over the **0** next to **Instances** and click the up arrow to increment the value to 1. Then click the **Scale** button. The status will go to **Scaling** and it will take some time for the app to start.

If all goes well the status will eventually go to **Running**. At this point the application is ready to be accessed. If the app doesn't start, review the logs for error messages. If this still doesn't help then please open a Support request.

# Login to Accelerator

To get the URL to connect to the web UI, click the **Copy URL** button in the middle of the application page to copy the value to the clipboard. 

Paste this into a new browser tab or window and append **/app/tracking/index.html** at the end of it. For example:

* **`eu-west-1.integration.cloud.tibcoapps.com/abc123/app/tracking/index.html`**

It is recommended to bookmark this page.

The browser will display the Login page for the Accelerator.

![Cloud Login](./images/cloud/login.png)

The credentials to login is an OAuth token generated from your TIBCO Cloud account. Click [here for EU region](https://eu.account.cloud.tibco.com/manage/settings/oAuthTokens) and [here for US region](https://account.cloud.tibco.com/manage/settings/oAuthTokens) to display the tokens page. Click the **Generate Token** button and given the token a name. Select a validity period. Beware that once the token expires you will need to generate another one in order to access the Accelerator. You must select the following domains:

* Connected Intelligence Cloud
* Spotfire®
* Data Streams
* Events
* Integration
* Messaging

![Generate OAUTH2 Token](./images/cloud/generate-token.png)

Note not all required domains are shown in the screenshot, scroll down for more.

Click the **Generate** button to generate the token. Note the confirmation dialog and then click the **Copy to clipboard** button to copy the token value.

![Token Granted](./images/cloud/token-granted.png)

Back on the Accelerator Login page paste the token value in the **OAuth Token** box and select the correct **Region** where your subscriptions are provisioned, then click **Login**.

If the credentials are valid then the Welcome page will be displayed. Both the OAuth token and Region will be stored in your browser local storage so will be saved for the next login. If you are using an incognito window or clear the browser cache the values will be lost, so it is worthwhile making a note of your OAuth token for future use.

# Generate Messaging Key

TIBCO Cloud Messaging comes preconfigured with a default role, so this step is optional. If you would like to make a specific role for the Accelerator click [here for EU region](https://eu.messaging.cloud.tibco.com/tcm/ui/admin) and [here for US region](https://messaging.cloud.tibco.com/tcm/ui/admin). Click the **Create Role** button. Enter a value for **Role Name** and keep the remaining options as defaults, then click the **Create** button. You do not need to note the role details as you will choose this later using a drop down.

# Configure Accelerator

The Welcome Page will be displayed on first load of the Accelerator and will walk you through the process of finalizing the provisioning of components.

## General Configuration

The first page will present some application options, adjust these as required.

* **Hide Welcome page** -- tick this box to hide the Welcome page on next login
* **Restrict UI** -- select the desired option for restricting access to the user interface. Note that your group will be different than the group **Accelerators** show in the screenshot.

Click the **Save** button to save any changes, and **Skip** to move to the next page without saving.

![Welcome Page 1](./images/cloud/welcome-1.png)

## TIBCO Cloud Messaging Configuration

The next page will present some TIBCO Cloud Messaging options. The **OAuth Token** will be preconfigured based on the one used to login. Select a **TIBCO Cloud Messaging Channel** from the drop down. This channel will be used for all Accelerator component communications. All your configured channels will be displayed, choose the desired option, if in doubt, select **default**. Click the **Save** button.

![Messaging Configuration](./images/cloud/welcome-2.png)

## Accelerator Component Provisioning

The next page presents component provisioning options. The **TIBCO Cloud Messaging Channel** will be displayed, as well as the **OAuth Token** used during login. You have the option of which components to provision. Select both **TIBCO Cloud Events** and **TIBCO Cloud Data Streams**. Click the **Provision** button. It will take some time for the provisioning to complete.

![Provisioning](./images/cloud/welcome-3.png)

Once the provisioning has completed you can verify the TIBCO Cloud Events application has been deployed by clicking [here for EU region](https://eu.events.cloud.tibco.com/) and [here for US region](https://events.cloud.tibco.com/) and check that you have a **track-eventmanager** application and it is either **Scaling** or **Running**. If the application is **Stopped**, hover over the status circle and click, then **Scale instances** to 1 to start.

Likewise you can verify TIBCO Cloud Data Streams by clicking [here for EU region](https://eu.datastreams.cloud.tibco.com/) and [here for EU region](https://datastreams.cloud.tibco.com/). You should have two streams called **Track_Parcel** and **Track_Route**. It will take several minutes for these to start so they may be in **Building** and **Starting** status for some time.

When all components have started correctly, back in the Welcome page you can click the **Next** button to continue.

## Spotfire® Configuration

The next page will display Spotfire® configuration options. Unfortunately it is not possible to automatically provision a Spotfire® DXP, so this will have to be done manually. The **OAuth Token** will be preconfigured based on the one used to login. Click the **Download DXP** button to dowload a file called **tracking.dxp** to your local machine that has been pre-configured to connect to the data streams previously provisioned.

Navigate to your TIBCO Cloud Spotfire subscription by clicking [here for EU region](https://eu.spotfire-next.cloud.tibco.com/) and [here for US region](https://spotfire-next.cloud.tibco.com/). Click the **Browse local file...** option and navigate to the **tracking.dxp** that you downloaded. Once loaded, the DXP should connect to your running streams showing an analytics dashboard. Everything will be empty for now though.

On the right hand side of the menu bar, where it says **Viewing**, change this to **Editing**. Then select the **File > Save As > Library Item...** menu option. Choose a location to save the analysis and then click the **Save** button.

Once it's saved you will need to extract the path to the analysis. The easiest way to do this is just extract it from the path in your browser. It will look something like this:

```
https://eu.spotfire-next.cloud.tibco.com/spotfire/wp/analysis?file=/Teams/XYZ4499/TrackTrace/tracking&waid=abc123&wavid=0
```
The part we are interested in here is the **file** query parameter. The value is between the "=" and the following "&". In the example above it will be:

```
/Teams/XYZ4499/TrackTrace/tracking
```

Back in the Accelerator Welcome page you can paste this in the **Analysis Path** box. The **Web Player Server** will be pre-populated based on the region that you logged in as, but please verify that it is correct by checking the URL of the Spotfire® page where you uploaded the DXP to ensure they match. Click the **Save** button once completed.

![Spotfire Configuration](./images/cloud/welcome-4.png)

The Welcome page settings are now completed so you can click **Close** to go to the main dashboard.

# Access Main Dashboard

The first time the Main Dashboard is accessed the UI will attempt to load the configured Spotfire® analysis. Note that third-party cookies must be enabled, so if you are using Incognito mode then this will need to be enabled by clicking the icon on the address bar, then **Site Not Working** and then **Allow Cookies**. You may need to manually refresh the page.

![Third Party Cookies](./images/cloud/third-party-cookies.png)

If you have not logged in to a TIBCO Cloud service in the current browser session, then dashboard will display a message prompting you to login to Spotfire®. Click the **Log in** button and complete the login process.

![Spotfire Login](./images/cloud/spotfire-login.png)

If you have previously logged in to a TIBCO Cloud service in the current browser session then the login step should be skipped and the dashboard will be displayed. It will be empty for now.

![Main Dashboard](./images/cloud/dashboard.png)

# Run a Simulation

Select the **Simulator** menu option. It may take a minute or two for the simulation status to become available to the webapp. Once the status is displayed, choose a value for **Select a testcase** from the list of options. The testcase name and a brief description are displayed. Then click the **Start Test Case** button.

![Simulator](./images/cloud/simulation.png)

The UI will change back to the dashboard view. A parcel will appear in the list on the left. Click on it mark the row and the remaining visualizations will be populated with relevant information.

![Execution](./images/cloud/execution.png)

# Adjust Settings

Select the **Settings** menu item to adjust any of the configurations defined during the Welcome page process. Note that components should not be re-provisioned without manually deleting any existing events applications or data streams. If data streams are re-provisioned then the Spotfire® DXP will require updating to point at the new streams.

![Settings](./images/cloud/settings.png)

# Shutting Down

There is a cost associated with running TIBCO Cloud Events, TIBCO Cloud Data Streams, and TIBCO Cloud Integration services. It is recommended to manually stop these services when not in use.

1. Open TIBCO Cloud Data Streams by clicking [here for EU region](https://eu.datastreams.cloud.tibco.com/) and [here for US region](https://datastreams.cloud.tibco.com/) then click the **Stop** button for each of **Track_Parcel**, and **Track_Route**. This will take a few minutes to stop.
1. Open TIBCO Cloud Integration by clicking [here for EU region](https://eu.integration.cloud.tibco.com/) and [here for US region](https://integration.cloud.tibco.com/) then then clicking the **Stop** button for the **track-webapp** application
1. Open TIBCO Cloud Events by clicking [here for EU region](https://eu.events.cloud.tibco.com/) and [here for US region](https://events.cloud.tibco.com/) then then clicking the **Stop** button for the **track-eventmanager** application
