/*
* Copyright © 2021. TIBCO Software Inc.
* This file is subject to the license terms contained
* in the license file that is distributed with this file.
*/

var axios = require('axios');
var fs = require('fs');
var archiver = require('archiver');
var properties = require('properties-reader');


// Environment properties
var cicProps = properties('./lib/cic.properties');

// Constants
var ZIP_FILENAME = "tracking.dxp";
var SPOTFIRE_BASE_PATH = __dirname + '/../data/provisioning/spotfire';

async function lookupStreams(region, authorization) {
    var promise = new Promise(function(resolve, reject){
        var url = '/api/v1/streams/all/runtime';
        var baseUrl = cicProps.get(region + '.TIBCO_CLOUD_DATASTREAMS');
      
        axios({
          url: url,
          method: 'GET',
          baseURL: baseUrl,
          headers: {
            'Authorization': authorization
          }
        })
        .then(function(response){
            resolve(response.data.streams);
        })
        .catch(function(err){
            reject(err);
        });
              
    });

    return promise;
};

async function writeDataAccessPlan(region, streams) {
    // Read the template file
    var dataAccessPlanTemplateFileName = SPOTFIRE_BASE_PATH + "/tracking/DataArchive/DataAccessPlan_Template.xml";
    var dataAccessPlanTemplate = fs.readFileSync(dataAccessPlanTemplateFileName, {encoding:'utf8', flag:'r'});

    var dataAccessPlan = dataAccessPlanTemplate;
    for(var thisStream of streams) {
        var url = cicProps.get(region + '.TIBCO_CLOUD_DATASTREAMS') + ':443/api/v1/streams/' + thisStream.id + '/runtime/connect'
        dataAccessPlan = dataAccessPlan.replace('${' + thisStream.param + '}', url);
    };

    // Update the data access plan
    var dataAccessPlanFileName = SPOTFIRE_BASE_PATH + "/tracking/DataArchive/DataAccessPlan.xml";
    fs.writeFileSync(dataAccessPlanFileName, dataAccessPlan);
};


async function download(region, authorization, res) {
    // Lookup the streams
    var streams = await lookupStreams(region, authorization);
    var streamMap = {};
    for(var thisStream of streams) {
        streamMap[thisStream.name] = thisStream;
    }

    // Expected streams
    var STREAMS = [
        {"name": "Track_Parcel", "param": "TCDS_PARCEL_URL"},
        {"name": "Track_Route", "param": "TCDS_ROUTE_URL"}
    ];
    
    // Map the id for the stream
    for(var thisStream of STREAMS) {
        var stream = streamMap[thisStream.name];
        if(stream == null) {
            res.status(500).send("Stream " + thisStream.name + " not found");
            return;
        }
        thisStream.id = stream.id;
    }
    
    await writeDataAccessPlan(region, STREAMS);

    // Create the zip
    var zip = archiver('zip');

    // Set the download name
    res.attachment(ZIP_FILENAME);

    await new Promise(function(resolve, reject) {
        // Pipe the archive to the response
        zip.pipe(res);

        // Append the directory of files
        zip.directory(__dirname + '/../data/provisioning/spotfire/tracking', false);

        // Error handler
        zip.on('error', function(err){
            throw err;
        });

        // Finalize the archive
        zip.finalize();

        // When the stream is finished
        res.on('close', function() {
            //res.send('OK').end();
            res.end();
            resolve();
        });
    }); 
};


module.exports.download = download;